#!
# cliSuppJimProc.tcl - CLI supporting TCL library.

# Copyright (c) 2006, TeamF1, Inc.

#
# modification history
# --------------------
# 01a,16jan06,smv written.

#
# DESCRIPTION
#
# This file containes procedures to support CLI.
#

# globals

set VAR_KEY_IDX      0
set VAR_KEY_VAL_IDX  1
set VAR_KEY_NAME_IDX 1

set VAR_ARG_NAME_IDX     0
set VAR_ARG_TYPE_IDX     1
set VAR_ARG_VALUES_IDX   2


###############################################################################
#
# cliExec - wrapper function to all TCL procedures.
#
# This routine is a wrapper function to all TCL procedures.
#
# RETURNS: 0 on success
#         -1 on error.
#
proc cliExec {argStr splitChars} {
    jimArgParser [jimArgSplit $argStr $splitChars]
}

###############################################################################
#
# jimArgSplit - performs argument parsing.
#
# This routine performs argument parsing. 
#
# RETURNS: array of tokens.
#
proc jimArgSplit {argLst splitChars} {
    set MyList ""

    #return if splitChars not specified or
    # splitChars contain only space
    if {([string length $splitChars] == 0) ||
        ([string match { } $splitChars] == 1)} {return $argLst}

    foreach arg $argLst {
        if {[string match {"*"} $arg] == 1} {
            lappend MyList [string range $arg 1 end-1]
            continue
            }
        if {[string match {*="*"} $arg] == 1} {
            set argLst [listTrim [split $arg =]]
            set res "[lindex $argLst 0]=[string range [lindex $argLst 1] 1 end-1]"
            if {[string match {* *} $res] == 1} {
                lappend MyList "\"$res\""
                } else {lappend MyList $res}
            continue
            }

        set argLen [string length $arg]
        if {[string match {/*} $arg] == 1} {
            lappend MyList /
            set arg [string range $arg 1 $argLen]
            }
        set endWithDir 0
        if {[string match {*/} $arg] == 1} {set endWithDir 1}
        # temporary fix to allow file paths as arguments only in key=value case.
        if {[string match {*=/*} $arg] != 1} {
            set tempRes [listTrim [split $arg /]]
            } else {set tempRes $arg}
        for {set i 0} {$i < [llength $tempRes]} {incr i} {
            if {$endWithDir || ([expr $i + 1] != [llength $tempRes])} {
                lappend MyList "[lindex $tempRes $i]/"
                } else {
                    if {[string match {* *} $[lindex $tempRes $i]] == 1} {
                        lappend MyList "\"[lindex $tempRes $i]\""
                        } else {lappend MyList "[lindex $tempRes $i]"}
                    }
            
            }
    }

    return $MyList
}

###############################################################################
#
# listTrim - creates a list based on the input
#
# This routine creates a list by taking non empty tokens in the given argument
# string.
#
# RETURNS: newly created list
#
proc listTrim {argStr} {
    set myLst ""
    foreach arg $argStr {
        if {[string length $arg] == 0} {continue}
        lappend myLst $arg
        }
    return $myLst
    }

###############################################################################
#
# argNameInfoGet - gets the argument information of the given parameter
#
# This routine returns the argument information of specified parameter
#
# RETURNS: key type.
#
proc argNameInfoGet {argInfoLst param} {
    global VAR_ARG_NAME_IDX
    set matched 0
    set argNameInfo ""
    foreach argInfo $argInfoLst {
        if {[string match {*,*} $argInfo] == 1} {
             set argInfo [listTrim [split $argInfo ,]]
             }
        set argNameInfo $argInfo
        set argInfo [listTrim [split [lindex $argInfo $VAR_ARG_NAME_IDX] :]]
        set nameIdx [expr [llength $argInfo] - 1]
        if {[string compare $param [lindex $argInfo $nameIdx]] == 0} {
            set matched 1
            break;
            }
        }
    if {$matched == 0} {set argNameInfo ""}
    return $argNameInfo
    }

    
###############################################################################
#
# argKeyInfoGet - gets name based in key value
#
# This routine returns the argument name of specified key by searching
# in the given keys list.
#
# RETURNS: key type.
#
proc argKeyInfoGet {argInfoLst param} {
    global VAR_KEY_IDX VAR_ARG_NAME_IDX
    set matched 0
    set argKeyInfo ""
    foreach argInfo $argInfoLst {
        if {[string match {*,*} $argInfo] == 1} {
             set argInfo [listTrim [split $argInfo ,]]
             }
        set argKeyInfo $argInfo
        set argInfo [listTrim [split [lindex $argInfo $VAR_ARG_NAME_IDX] :]]
        if {[llength $argInfo] == 1} {continue}
        if {[string compare $param [lindex $argInfo $VAR_KEY_IDX]] == 0} {
            set matched 1
            break;
            }
        }
    if {$matched == 0} {set argKeyInfo ""}
    return $argKeyInfo
    }

set DISPLAY_TITLE_FLAG_IDX  0
set DISPLAY_ALT_TITLE_IDX   1
set DISPLAY_KEYS_IDX        2


proc formatTokensGet {frmtStr} {
    set fmtSpec "%"
    set fmtTokLst ""
    if {[string length $frmtStr] > 1 } {
        set fmtLst [listTrim [split $frmtStr %]]
        if {[llength $fmtLst] == 1} {return $frmtStr}
        if {[string match {%*} $frmtStr] == 0} {
            lappend fmtTokLst "[lindex $fmtLst 0]%[lindex $fmtLst 1]"
             } else {
             lappend fmtTokLst "%[lindex $fmtLst 0]"
             lappend fmtTokLst "%[lindex $fmtLst 1]"
             }
             for {set i 2} {$i < [llength $fmtLst]} {incr i} {
                 lappend fmtTokLst "%[lindex $fmtLst $i]"
                 }
        }
#    listShow $fmtTokLst "fmtTokLst"
    return $fmtTokLst
    }

proc keyNamePairLstGet {keyInfo titleInfo formatStr} {
    global VAR_KEY_IDX VAR_ARG_NAME_IDX
    set keyNameLst ""
    if {[string length $keyInfo] > 0 } {
         set keyInfo [listTrim [split $keyInfo \{*\}]]
         } else {return $keyNameLst}
    
    if {[string length $titleInfo] <= 0 } {return $keyNameLst}
    set titleRow [listTrim [split [listTrim [split $titleInfo]] {<> }]]
    
    set formatStrLst [formatTokensGet $formatStr]
    if {[llength $formatStrLst] != [llength $titleRow]} {return ""}
    
    for {set i 0} {$i < [llength $formatStrLst]} {incr i} {
        set formatSpec [lindex $formatStrLst $i]
        set argName [lindex $titleRow $i]
        set argInfo [lindex [argNameInfoGet $keyInfo $argName] $VAR_ARG_NAME_IDX]
        if {[string length $argInfo] == 0} {return ""}
        set argInfo [listTrim [split [lindex $argInfo $VAR_ARG_NAME_IDX] :]]
        set keyVal $argInfo
        if {[llength $argInfo] == 2} {
            set keyVal [lindex $argInfo $VAR_KEY_IDX]
            }
        lappend keyNameLst "$formatSpec" $keyVal
        }
#    listShow $keyNameLst "keyNameLst "
    return $keyNameLst
    }
    
###############################################################################
#
# cmdSyntaxShow - displays command syntax
#
# This routine displayes command syntax after stripping all unnecessary user
# information.
#
# RETURNS: syntax string.
#
proc cmdSyntaxGet {grpArgsInfoStr syntaxStr} {
    global VAR_ARG_NAME_IDX VAR_ARG_TYPE_IDX VAR_ARG_VALUES_IDX
    global VAR_KEY_NAME_IDX
    if {[llength $grpArgsInfoStr] > 0 } {
         set grpArgsInfoStr [listTrim [split $grpArgsInfoStr \{*\}]]
         }
#    puts "syntaxStr : $syntaxStr "
    if {[string length $syntaxStr] == 0 } {return ""}
    set syntaxStr [listTrim [split $syntaxStr ]]
    # only command name
    if {[llength $syntaxStr] == 1 } {return $syntaxStr}
    # store command name in returning list
    set syntax [lindex $syntaxStr 0]
    # skip first values as it is command name
    set syntaxStr [lrange $syntaxStr 1 [llength $syntaxStr]]

    set paramLen 0
    # for each parameter in syntax extract parameter name and type
    foreach paramStr $syntaxStr {
        set optArg 0
        set paramLst ""

        if {[string match {\[*\]} $paramStr] == 1} {
            set paramLst [listTrim [split $paramStr \[,\]]]
            set optArg 1
            } elseif {[string match {<*>} $paramStr] == 1} {
                set paramLst [listTrim [split $paramStr <,>]]
                } else {
                    # constant argument
                    lappend syntax $paramStr
                    continue
                    }

       set argName ""
       set argType ""
       set validVals ""
      
       # extract argument type from paramLst string
       # {NAME TYPE VALIDRANGE}
       if {[llength $paramLst] > 1} {
           if {[llength $paramLst] > 2} {
               set validVals [lindex $paramLst $VAR_ARG_VALUES_IDX]
               }
           set argType [lindex $paramLst $VAR_ARG_TYPE_IDX]
           if {[string match {ARG_*} $argType] == 0} {
               set argType ""
               set validVals [lindex $paramLst $VAR_ARG_VALUES_IDX]
               }
           set argName [lindex $paramLst $VAR_ARG_NAME_IDX]
           } else {set argName $paramLst}

        # if syntax does not contain argument type information get it from group
        # commands information (if specified)
        if {([string length $argType] == 0) &&
            ([string length $grpArgsInfoStr] > 0)} {
                
             set argInfo [argNameInfoGet $grpArgsInfoStr $argName]
             if {[string length $argInfo] == 0} {return ""}
             # if cliCmdKey is not specified, extract it from argument
             # information.
             set argName [lindex $argInfo $VAR_ARG_NAME_IDX]
             # if argName is in key:name format, then split it and
             # extract values.
             if {[string match {*:*} $argName] == 1} {
                 set argName [split $argName :]
                 if {[llength $argName] != 2} {
                     puts "invalid argument name - '$argName'"
                     return ""
                     }
                  set argName [lindex $argName $VAR_KEY_NAME_IDX]
                  }
             # get argument type
             set argType [lindex $argInfo $VAR_ARG_TYPE_IDX]
             }

        if {[string length $argType] > 0 } {
            set param "$argName - [string range $argType 4 [string length $argType]]"
            } else {set param $argName}

        if {[string length $validVals]} {
            lappend param "-"
            set validVals [split $validVals \;]
            foreach value $validVals {
                if {[string match {*:*} $value]} {
                    set value [lindex [split $value :] 0]
                    }
                lappend param $value
                }
            }

        if {$optArg} {
            lappend syntax "\[$param\]" 
            } else {lappend syntax "<$param>"}
        }

    return $syntax;
}

###############################################################################
#
# cmdSynCheckWrapper - wapper around cmdSynCheck
#
# This routine is a wrapper function around cmdSynCheck. This function takes
# command syntax and tokenizes it based on a specific predefined format and
# calls cmdSynCheck.
#
# RETURNS: key,value pair list.
#
proc cmdSynCheckWrapper {keyInfoStr syntax argStr mgmtFlag} {
    if {[llength $keyInfoStr] > 0 } {
         set keyInfoStr [listTrim [split $keyInfoStr \{*\}]]
         }
    if {$mgmtFlag == 1} {
        return [cmdSynCheck1 $keyInfoStr $syntax $argStr]
        } else {
             return [cmdSynCheck $keyInfoStr $syntax $argStr]
             }
}

###############################################################################
#
# cliCmdKeySet - preapres key=value pair and append to existing list.
#
# This routine preapres key=value pair and append to existing list.
#
# RETURNS: key,value pair list.
#
proc cliCmdKeySet {key val} {
    if {[string length $key] == 0} {
        return $val
        } else {return "$key=$val"}
    }

###############################################################################
#
# cmdSynCheck - performs command syntax checking and argument type checking.
#
# This routine matches the parameters against the command syntax. It performs
# argument type checking and validates the argument against given set of values.
# All the validating information should be specified in the command syntax
# string. This routine expects command string as
#    command <manArg1, type, valid range of values>
#            [optArg1, type, valid range of values>
# The argument information enclosed in < and > are treated as mandatory
# arguments and argument information enclosed in [ and ] are treated as optional
# arguments. 
# The valid values for 'type' information are
#    ARG_INT     -  integer
#    ARG_STRING  -  character array / string
#    
#    ARG_IPV4_ADDR - ipv4 address.
# The 'valid range of values' can be in two forms.
#     1. values seperated with a '-'
#     2. values seperated with a ':'.
# The first method generally used to specify a range for integer parameters. 
# for ex if we want to check port number, the port number should be between
# 0 and 65535. It can be specified in the syntax as
#    <port, ARG_INT, 0-65535>
# The second method generally used when we want to check the argument against
# different set of values. For example, the valid values for the protocol feild
# are ip, icmp, tcp, udp, etc.. we can specify this in the command syntax as
#   <protocol, ARG_STRING, any:ip:icmp:tcp:udp>
#
# RETURNS: key,value pair list.
#
proc cmdSynCheck {argsInfoStr cmdSyntax argStr} {
    # map required global variables
    global VAR_KEY_IDX VAR_KEY_NAME_IDX VAR_KEY_VAL_IDX
    global VAR_ARG_NAME_IDX VAR_ARG_TYPE_IDX VAR_ARG_VALUES_IDX
    # create list to store {key=value} pairs for valid arguments
    set keyValLst ""
    set cmdSyntax [listTrim [split $cmdSyntax]]
    # strip first values as it is command name
    set cmdSyntax [lrange $cmdSyntax 1 [llength $cmdSyntax]]
    # count number of expected arguments specified in syntax string
    set nCmdArgs [llength $cmdSyntax]
#    listShow $cmdSyntax "cmdSyntax"

    # Ignore all input arguments if command does not expect arguments
    if {$nCmdArgs == 0} {return 0} 

#    listShow  $argStr "argStr"
    set nInputArgs [llength $argStr]

    # check number of args if syntax doesnot contains indicator for variable
    # number of arguments.
    if {[string match {*...*} $cmdSyntax] != 1} {
         # number of given arguments should not be greater than expected
         # number of arguments.
         if {$nInputArgs > $nCmdArgs} {
             puts "Invalid number of arguments(more)"
             return -1
            }
        }
    # count number of optional arguments in syntax
    set nOptArgs 0
    foreach param $cmdSyntax {
        if {[string match {\[*\]} $param] == 1} {incr nOptArgs}
        }

    if {($nCmdArgs == $nOptArgs) && ($nInputArgs == 0)} {return 0}

    # cosider non iptional arguments as mandatory arguments
    # i.e. mandatory + constant arguments
    set nManArgs [expr $nCmdArgs - $nOptArgs]

    # input arguments should not be less than required arguments
    if {$nInputArgs < $nManArgs} {
        puts "Invalid number of arguments(less)"
        return -1
        }

    # perform type checking and value checking for each argument.
    while {$nInputArgs > 0} {

       # check number of args if syntax doesnot contains indicator for variable
       # number of arguments.
       if {($nCmdArgs == 1) &&
           ([string match {*...*} $cmdSyntax] == 1)} {
           foreach arg $argStr {
               lappend keyValLst $arg
               }
           break
           }

       set cliCmdKeyGiven 0
       set cliCmdKey ""

       # get first argument from list
       set argVal [lindex $argStr 0]
       # if argument is in key=value format then strip 'key=' portion
       set argVal [split $argVal =]
       if {[llength $argVal] > 1} {
           if {[llength $argVal] >= 2} {
               set cliCmdKeyGiven 1
               set cliCmdKey [lindex $argVal $VAR_KEY_IDX]
               set argVal [lindex $argVal $VAR_KEY_VAL_IDX]
               } else {
                   puts "Invalid argument format"
                   return -1
                   }
               }
      
       set paramLst ""
       set argCheckedTypes ""
       # check for first parameter match in optional or mandaotry or constant
       # parameters.
       while {$nCmdArgs > 0} {
           set optArg 0
#           set skipParam 0
         
           # extract first parameter from list
           set paramStr [lindex $cmdSyntax 0]
           # decrement number of parameter in command string
           set nCmdArgs [expr $nCmdArgs - 1]
           # strip this parameter from list
           set cmdSyntax [lrange $cmdSyntax 1 $nCmdArgs]

#           puts "paramStr: $paramStr"

           # optional/mandatory ??
           if {[string match {\[*\]} $paramStr] == 1} {
               set nOptArgs [expr $nOptArgs - 1]
               set paramLst [listTrim [split $paramStr \[,\]]]
               if {$nInputArgs <= $nManArgs} {
#                    set skipParam 1
                    # skip and strip this parameter as we have more constant or
                    # mandatory parameters to check.
                    if {$cliCmdKeyGiven} {return -1}
                    set cliCmdKey ""
                    continue
                    } else {
                        set optArg 1
                      }
               } elseif {[string match {<*>} $paramStr] == 1} {
                     set paramLst [listTrim [split $paramStr <,>]]
                     } else {
                         # check for constant parameter
                         if {[string compare $argVal $paramLst] != 0} {
                             puts "invalid constant argument '$argVal' expecting '$paramLst'"
                             return - 1
                             }
                         if {$cliCmdKeyGiven} {return -1}
                         set cliCmdKey ""
                         continue
                         }
#           listShow $paramLst "paramLst"

           set argName ""
           set argType ""
           set validRange ""

           # get first argument name and check whether it is in key:name
           # format or not. If it is in key:name format, then split it in to
           # key, name. If user specified argument in key=value format then 
           # compare given keyValue with this key to verify the correctness
           # of argument.
           set argName [lindex $paramLst $VAR_ARG_NAME_IDX]
#           puts "argName: $argName"
           if {[string match {*:*} $argName] == 1} {
               set argNameInfo [split $argName :]
               if {[llength $argNameInfo] != 2} {
                   puts "invalid argument name - '$argName'"
                   return -1
                   }
               if {[string length $cliCmdKey] > 0]} {
                   set argNameVal [lindex $argNameInfo $VAR_KEY_IDX]
                   if {$cliCmdKey != $argNameVal} {
                     if {$nCmdArgs == 0 || $optArg == 0} {
                         puts "invalid keyValue. expecting '$argNameVal' as it is mandatory"
                         return -1
                         }
                     if {$cliCmdKeyGiven} {return -1}
                     set cliCmdKey ""
                     continue
                     }
                   }
               set argName [lindex $argNameInfo $VAR_KEY_NAME_IDX]
               }
#           puts "argName: $argName"

           # extract argument type from paramInfo string
           # {NAME TYPE VALIDRANGE}
           if {[llength $paramLst] > 1} {
               if {[llength $paramLst] > 2} {
                   set validRange [lindex $paramLst $VAR_ARG_VALUES_IDX]
                   }
               set argType [lindex $paramLst $VAR_ARG_TYPE_IDX]
               if {[string match {ARG_*} $argType] == 0} {
                   set validRange [lindex $paramLst $VAR_ARG_VALUES_IDX]
                   }
               }
#            puts "argType: '$argType' range: '$validRange'"

#           puts "argsInfoStr: $argsInfoStr"
           
           if {([llength $argsInfoStr] != 0) && 
               (([string length $argType] == 0) ||
                ([string length $validRange] == 0))} {
               set argInfo ""
#               puts "GOING TO FIND ARGUMENT INFORMATION"
#               puts "cliCmdKey: '$cliCmdKey'"
               # arument is given in key=value format, then get argument
               # information based on key value. otherwise get argument
               # information based on command name
               if {[string length $cliCmdKey] > 0} {
                   set argInfo [argKeyInfoGet $argsInfoStr $cliCmdKey]
                   } else {
                       set argInfo [argNameInfoGet $argsInfoStr $argName]
                       }
               if {[string length $argInfo] == 0} {return -1}
               
#               puts "argInfo : $argInfo "
               # if cliCmdKey is not specified, extract it from argument
               # information.
               if {[string length $cliCmdKey] == 0} {
                   set argName [lindex $argInfo $VAR_ARG_NAME_IDX]
                   # if argName is in key:name format, then split it and
                   # extract values.
                   if {[string match {*:*} $argName] == 1} {
                       set argName [split $argName :]
                       if {[llength $argName] != 2} {
                           puts "invalid argument name - '$argName'"
                           return -1
                           }
                       set cliCmdKey [lindex $argName $VAR_KEY_IDX]
                       set argName [lindex $argName $VAR_KEY_NAME_IDX]
                       }
                   }
               # if argType is not specified, extract it from argument
               # information.
               if {[string length $argType] == 0} {
                   set argType [lindex $argInfo $VAR_ARG_TYPE_IDX]
                   }
               # if validRange is not specified, extract it from argument
               # information.
               if {[string length $validRange] == 0} {
                   set validRange [lindex $paramLst $VAR_ARG_VALUES_IDX]
                   }
               }
                
           # update number of mandatory arguments
           set nManArgs [expr $nCmdArgs - $nOptArgs]
#           puts "argName: $argName argType: $argType validRange: $validRange"
           if {[string length $argType] > 0} {
               lappend argCheckedTypes "($argName)$argType"
               # perform type checking.
               if {[argValidate $argType $argVal] != 0} {
                   if {$nCmdArgs == 0 || $optArg != 1} {
                       puts "Invalid $argName: $argVal"
                       return -1
                       }
                   lappend argCheckedTypes "||"
#                   puts "skip '$argVal' Check with '$argName - $argType'"
                   if {$cliCmdKeyGiven} {return -1}
                   set cliCmdKey ""
                   continue
                   }
               }
          
           # perform value checking.
           if {[string length $validRange] > 0} {
               set matchedArg [argRangeValidate $argType $argVal $validRange]
#               puts "matchedArg: '$matchedArg' ([string length $matchedArg])"
               if {[string length $matchedArg] == 0} {
                   if {$nCmdArgs == 0 || $optArg !=1} {
                       if {[llength $argCheckedTypes] > 0} {
                           puts "Invalid $argName: $argVal"
                           puts "and"
                           }
                       puts "'$argVal' is not in specified range - '($argName)$validRange'"
                       return -1
                       }
                   # skip this parameter as we have some more to check
                   if {$cliCmdKeyGiven} {return -1}
                   set cliCmdKey ""
                   continue
                   } else {set argVal $matchedArg}
               }

           # found a match, store information and break from inne while loop.
#          puts "appending $cliCmdKey - $argVal"
          lappend keyValLst [cliCmdKeySet $cliCmdKey $argVal]
          break
          }

         # check number of args if syntax doesnot contains indicator for
         # variable number of arguments.
         if {[string match {*...*} $cmdSyntax] != 1} {
             if {$nInputArgs > [expr $nCmdArgs + 1]} {
                 puts "invalid arguments"
                 return -1
                 }
             }

        # decrement number of input arguments
        set nInputArgs [expr $nInputArgs - 1] 
        # strip first argument from list
        set argStr [lrange $argStr 1 $nInputArgs]
       }

#   #listShow $keyValLst "keyValLst"
   return $keyValLst; 
}

proc cmdSynCheck1 {argsInfoStr cmdSyntax argStr} {
    # map required global variables
    global VAR_KEY_IDX VAR_KEY_NAME_IDX VAR_KEY_VAL_IDX
    global VAR_ARG_NAME_IDX VAR_ARG_TYPE_IDX VAR_ARG_VALUES_IDX
    # create list to store {key=value} pairs for valid arguments
    
    set keyValLst ""
    set cmdSyntax [listTrim [split $cmdSyntax]]
    # strip first values as it is command name
    set cmdSyntax [lrange $cmdSyntax 1 [llength $cmdSyntax]]
    # count number of expected arguments specified in syntax string
    set nCmdArgs [llength $cmdSyntax]
#    listShow $cmdSyntax "cmdSyntax"

    # Ignore all input arguments if command does not expect arguments
    if {$nCmdArgs == 0} {return 0} 

#    listShow  $argStr "argStr"
    set nInputArgs [llength $argStr]

    # check number of args if syntax doesnot contains indicator for variable
    # number of arguments.
    if {[string match {*...*} $cmdSyntax] != 1} {
         # number of given arguments should not be greater than expected
         # number of arguments.
         if {$nInputArgs > $nCmdArgs * 2} {
             if {[string match {*=*} [lindex $argStr 0]] == 0} {
                 puts "Invalid number of arguments(more)"
                 return -1
                 }
            }
        }
        
    set cliCmdKeyGiven 0

    # perform type checking and value checking for each argument.
    while {$nInputArgs > 0} {
       set cliCmdKey ""

       # get first argument from list
       set argVal [lindex $argStr 0]
#       puts "inputArg: $argVal"
       # if argument is in key=value format then strip 'key=' portion
       set argVal [split $argVal =]
       if {[llength $argVal] > 1} {
           if {[llength $argVal] >= 2} {
               set cliCmdKeyGiven 1
               # based on discussion with team lead, we are skipping type and
               # value checking if the input is in key=value format. To skip
               # this just comment below lines upto continue and uncomment next
               # lines.
               lappend keyValLst [lindex $argStr 0]
               # decrement number of input arguments
               set nInputArgs [expr $nInputArgs - 1] 
               # strip first argument from list
               set argStr [lrange $argStr 1 $nInputArgs]
               continue
                
               # uncomment below two lines if you want to have type and value
               # checkings.
#               set cliCmdKey [lindex $argVal $VAR_KEY_IDX]
#               set argVal [lindex $argVal $VAR_KEY_VAL_IDX]

               } else {
                   puts "Invalid argument format"
                   return -1
                   }
           } elseif {$cliCmdKeyGiven == 1} {
               # we already got key=value pair, but it is not consistent
               # trough out input arguments
               puts "all input values must be in identical format"
               return -1
               }
#       puts "argVal: $argVal"

       set paramInfo ""
       set tmpInfoStr ""
       if {[string length $argsInfoStr] > 0} {
           set tmpInfoStr $argsInfoStr
           } else {set tmpInfoStr $cmdSyntax}
           
       # If input is in key=value format, then find parameter information
       # based on key value. otherwise find parameter information based on the
       # given argument name
       if {$cliCmdKeyGiven == 1} {
           set paramInfo [argKeyInfoGet $tmpInfoStr $cliCmdKey]
           if {[string length $paramInfo] == 0} {
               puts "invalid input key - $cliCmdKey"
               return -1
               }
           } else {
               set paramInfo [argNameInfoGet $tmpInfoStr $argVal]
               if {[string length $paramInfo] == 0} {
                   puts "invalid input argument - $argVal"
                   return -1
                   }
               }
#        puts "Result - $cliCmdKey:$argVal - $paramInfo"

        if {$cliCmdKeyGiven == 0} {
            # extract next argument
            set argVal [lindex $argStr 1]

            # decrement number of input arguments
            set nInputArgs [expr $nInputArgs - 1] 
            # strip first argument from list
            set argStr [lrange $argStr 1 $nInputArgs]
            }

        if {[llength $paramInfo] > 1 } {
            if {[argTypeAndValueCheck $paramInfo $argVal] == 0} {return -1;}
            }
        
        if {$cliCmdKeyGiven == 0} {
            set argName [lindex $paramInfo $VAR_ARG_NAME_IDX]
            if {[string match {*:*} $argName] == 1} {
                set argNameInfo [split $argName :]
                set cliCmdKey [lindex $argNameInfo $VAR_KEY_IDX]
                }
            }

        lappend keyValLst [cliCmdKeySet $cliCmdKey $argVal]

        # decrement number of input arguments
        set nInputArgs [expr $nInputArgs - 1] 
        # strip first argument from list
        set argStr [lrange $argStr 1 $nInputArgs]
        }
        
    if {[string length $keyValLst] == 0} {return 0}
    #listShow $keyValLst "keyValLst"
    return $keyValLst; 
}

proc argTypeAndValueCheck {paramInfo argVal} {
    global VAR_KEY_IDX VAR_KEY_NAME_IDX
    global VAR_ARG_NAME_IDX VAR_ARG_TYPE_IDX VAR_ARG_VALUES_IDX

    set argName ""
    set argType ""
    set validRange ""

    set argName [lindex $paramInfo $VAR_ARG_NAME_IDX]
    if {[string match {*:*} $argName] == 1} {
        set argNameInfo [split $argName :]
        if {[llength $argNameInfo] != 2} {
            puts "invalid argument name - '$argName'"
            return 0
            }
        set argName [lindex $argNameInfo $VAR_KEY_NAME_IDX]
        }

    # extract argument type from paramInfo string
    # {NAME TYPE VALIDRANGE}
    if {[llength $paramInfo] > 1} {
        if {[llength $paramInfo] > 2} {
            set validRange [lindex $paramInfo $VAR_ARG_VALUES_IDX]
            }
        set argType [lindex $paramInfo $VAR_ARG_TYPE_IDX]
        if {[string match {ARG_*} $argType] == 0} {
            set validRange [lindex $paramInfo $VAR_ARG_VALUES_IDX]
            }
        }

#    puts "argValue: '$argVal'"
#    puts "argName: '$argName' argType: '$argType' validRange: '$validRange'"
     if {[string length $argType] > 0} {
         # perform type checking.
         if {[argValidate $argType $argVal] != 0} {
             puts "Invalid type $argName: $argVal"
             return 0
             }
         }
          
     # perform value checking.
     if {[string length $validRange] > 0} {
         set matchedArg [argRangeValidate $argType $argVal $validRange]
#        puts "matchedArg: '$matchedArg' ([string length $matchedArg])"
         if {[string length $matchedArg] == 0} {
             puts "Invalid value for $argName : $argVal"
             return 0
             }
        }

    return 1
    }

###############################################################################
#
# argValidate - performs argument type checking.
#
# This routine performs argument type checking.
# 
# Currently this routine performs type checking for the following types.
# Primitive datatypes
#    - Integers
#
# Compound datatypes
#    - IPV4 Address.
#
# RETURNS: 0 on success
#         -1 on error.

proc argValidate {argType argStr} {
#     puts "argValidate: $argStr - $argType"
     switch $argType {
         ARG_BOOL {
              if {[boolTypeCheck $argStr] != 0} {return -1}
         }
         ARG_INT {
              if {[intTypeCheck $argStr] != 0} {return -1}
         }
         ARG_STRING {
         }
         ARG_IPV4ADDR {
             if {[ipv4AddrFormatCheck $argStr] != 0 } {return -1}
         }
         ARG_MAC {
             if {[macAddrFormatCheck $argStr] != 0 } {return -1}
         }
         default {
         }
     }
    return 0;
}

###############################################################################
#
# argRangeValidate - performs checking againest given parameter values
#
# This routine checks whether given argument value is in the specified range or
# not.
#
# RETURNS: 0 on success
#         -1 on error.

proc argRangeValidate {argType argValue argValueRange} {
#    puts "argRangeValidate: $argType $argValue $argValueRange"
    set checkType [rangeChekingTypeGet $argValueRange]
    switch $argType {
        ARG_INT {
            return [intValidRangeCheck $argValue $checkType $argValueRange]
        }
        ARG_STRING {
            return [strValidRangeCheck $argValue $checkType $argValueRange]
        }
        ARG_IPADDR {
        }
        default {
        }
    }
    return 0;
}

###############################################################################
#
# ipv4AddrFormatCheck - validates IPV4 address format.
#
# This routine checks whether given argument is a valid IPV4 address or not.
#
# RETURNS: 0 on success
#         -1 on error.
proc ipv4AddrFormatCheck {ipAddr} {
    set ipAddrBytes [split $ipAddr .]
    if {[llength $ipAddrBytes] != 4} {return -1}
    foreach byte $ipAddrBytes {
       if {[string length $byte] == 0 || [string length $byte] > 3} {return -1}
       if {[intTypeCheck $byte] != 0} {return -1}
       if {$byte > 255} {return -1}
       }
    return 0;
}

###############################################################################
#
# macAddrFormatCheck - checks mac address format.
#
# This routine checks whether given argument is a valid mac address or not.
#
# RETURNS: 0 on success
#         -1 on error.
proc macAddrFormatCheck {macAddr} {
    set macAddrBytes [split $macAddr :]
    if {[llength $macAddrBytes] != 6} {return -1}
    foreach byte $macAddrBytes {
       if {[string match {0[xX]*} $byte] == 1} {
           set byte [string range $byte 2 [string length $byte]]
           }
       if {[llength $byte] > 2} {return -1}
       if {[string length $byte] == 1} {
           if {[string match {[0-9a-fA-F]} $byte] == 0} {return -1}
           } else {
              if {[string match {[0-9a-fA-F][0-9a-fA-F]} $byte] == 0} {
                  return -1
                  }
              }
       }
    return 0;
}

###############################################################################
#
# intTypeCheck - validates integer.
#
# This routine ensures that given argument contains only numeric numbers in it.
#
# LIMITATION:
#    As JIM does not support regular expressions, this routine uses globbing
#    wildcards for type identification.
#
# RETURNS: 0 on success
#         -1 on error.
#
proc intTypeCheck {arg} {
    set pattern {*[a-zA-z,<.>/?:;\"'\[\]\{\}~`!@#$%^&*()_+\\|]*}
    if {[string match $pattern $arg] == 1} {return -1}
    return 0;
}

###############################################################################
#
# boolTypeCheck - validates boolean.
#
# This routine ensures that given argument contains only boolean values.
#
# LIMITATION:
#    As JIM does not support regular expressions, this routine uses globbing
#    wildcards for type identification.
#
# RETURNS: 0 on success
#         -1 on error.
#
proc boolTypeCheck {arg} {
    puts "Given Arg: [string toupper $arg]"
    switch [string toupper $arg] {
        "0"      {return 0}
        "1"      {return 0}
        "YES"    {return 0}
        "NO"     {return 0}
        "ON"     {return 0}
        "OFF"    {return 0}
        "TRUE"   {return 0}
        "FALSE"  {return 0}
        }
    
    return -1
}

###############################################################################
#
# intValidRangeCheck - validates argument value.
#
# This routine ensures that given argument is in specified range.
#
# RETURNS: 0 on success
#         -1 on error.
proc intValidRangeCheck {arg checkType valueStr} {
#    puts "strValidRangeCheck: $arg $checkType $valueStr"
    set matchedArg $arg
    switch $checkType {
        RANGE_CHECK {
            set minMaxVals [split $valueStr -]
            if {[llength $minMaxVals] > 2} {return ""}
            if {$arg < [lindex $minMaxVals 0] || $arg > [lindex $minMaxVals 1]} {
                return ""
                }
        }
        MULTI_VALUES {
        }
        NONE {
        }
        default {
        }
    }
             
#    puts "returning: $matchedArg"
    return $matchedArg;
}

###############################################################################
#
# strValidRangeCheck - validates argument value.
#
# This routine ensures that given argument is in specified range.
#
# RETURNS: 0 on success
#         -1 on error.
proc strValidRangeCheck {arg checkType valueStr} {
#    puts "strValidRangeCheck: $arg $checkType $valueStr"
    set matchedArg $arg
    switch $checkType {
        RANGE_CHECK {
            set valLst [listTrim [split $valueStr -]]
            if {[llength $valLst] > 1} {
                puts "not implemented"
                }
            if {[string compare $arg $valueStr] != 0} {return ""}
        }
        MULTI_VALUES {
            set valLst [listTrim [split $valueStr \;]]
            set matchFound 0
            foreach val $valLst {
               set optVal ""
               if {[string match {*:*} $val]} {
                   set val [split $val :]
                   set optVal [lindex $val 1]
                   set val [lindex  $val 0]
                   }
               if {[string compare $val $arg] == 0} {
                   set matchFound 1
                   if {[string length $optVal]} {
                       set matchedArg $optVal
                       }
                   break
                }
            }
            if {$matchFound == 0} {return ""}
        }
        NONE {
        }
        default {
            if {[string compare $arg $valueStr] != 0} {return ""}
        }
    }
#    return "returning: $matchedArg"
    return $matchedArg
}

###############################################################################
#
# rangeChekingTypeGet - derives the range specification type from value string.
#
# This routine derives the range specification type from value string.
#
# RETURNS: range type on success
#          0 on error.
proc rangeChekingTypeGet {valueStr} {
    if {[string compare "NONE" $valueStr] == 0} {return "NONE"}
    if {[string match {*;*} $valueStr] != 0} {return "MULTI_VALUES"}
    if {[string match {*-*} $valueStr] != 0} {return "RANGE_CHECK"}
    return 0;
}

proc listShow {lst descr} {
    puts "----------- $descr (Len: $[llength $lst]) ------------"
    foreach str $lst {puts "\t'$str'"}
    puts "------------------------------------------------------"
}
